# -*- coding: utf-8 -*-
"""
Class for transmitting content metadata to Degreed.
"""

from integrated_channels.degreed2.client import Degreed2APIClient
from integrated_channels.integrated_channel.transmitters.content_metadata import ContentMetadataTransmitter


class Degreed2ContentMetadataTransmitter(ContentMetadataTransmitter):
    """
    This transmitter transmits exported content metadata to Degreed2.
    """

    def __init__(self, enterprise_configuration, client=Degreed2APIClient):
        """
        Use the ``Degreed2APIClient`` for content metadata transmission to Degreed.
        """
        super().__init__(
            enterprise_configuration=enterprise_configuration,
            client=client
        )

    def _prepare_items_for_transmission(self, channel_metadata_items):
        # similar to canvas, we can't create courses in bulk hence limiting to size 1
        # this of course only is accurate if transmission chunk size is 1
        return {
            'courses': [channel_metadata_items[0]],
        }



# Generated by Django 3.2.11 on 2022-02-09 14:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('blackboard', '0013_blacboardglobalconfiguration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='blackboardlearnerassessmentdatatransmissionaudit',
            name='enterprise_course_enrollment_id',
            field=models.IntegerField(db_index=True),
        ),
    ]



# Generated by Django 2.2.24 on 2021-07-16 18:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sap_success_factors', '0004_auto_20210708_1639'),
    ]

    operations = [
        migrations.AddField(
            model_name='sapsuccessfactorsenterprisecustomerconfiguration',
            name='prevent_self_submit_grades',
            field=models.BooleanField(default=False, help_text="When set to True, the integration will use the generic edX service user ('sapsf_user_id') defined in the SAP Customer Configuration for course completion.", verbose_name='Prevent Learner From Self-Submitting Grades'),
        ),
    ]



"""
Package for generic data transmitters which send serialized data to integrated channels.
"""


class Transmitter:
    """
    Interface for transmitting data to an integrated channel.

    The interface contains the following method(s):

    transmit(payload)
        payload - The ``Exporter`` object expected to implement an ``export`` method that returns serialized data.
    """

    def __init__(self, enterprise_configuration, client=None):
        """
        Prepares a configuration and a client to be used to transmit data to an integrated channel.

        Arguments:
            * enterprise_configuration - The configuration connecting an enterprise to an integrated channel.
            * client - The REST API client that'll transmit serialized data.
        """
        self.enterprise_configuration = enterprise_configuration
        self.client = client(enterprise_configuration) if client else None

    def transmit(self, create_payload, update_payload, delete_payload, content_updated_mapping, **kwargs):
        """
        The abstract interface method for sending exported data to an integrated channel through its API client.
        """
        raise NotImplementedError('Implement in concrete subclass transmitter.')



# Generated by Django 3.2.11 on 2022-02-03 17:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sap_success_factors', '0010_sapsuccessfactorsenterprisecustomerconfiguration_display_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sapsuccessfactorslearnerdatatransmissionaudit',
            name='enterprise_course_enrollment_id',
            field=models.IntegerField(db_index=True),
        ),
    ]



"""
Client for connecting to Cornerstone.
"""

import base64
import json
import logging

import requests

from django.apps import apps

from integrated_channels.cornerstone.utils import get_or_create_key_pair
from integrated_channels.integrated_channel.client import IntegratedChannelApiClient
from integrated_channels.utils import generate_formatted_log

LOGGER = logging.getLogger(__name__)


class CornerstoneAPIClient(IntegratedChannelApiClient):
    """
    Client for connecting to Cornerstone.

    Specifically, this class supports obtaining access tokens
    and posting user's course completion status to progress endpoints.
    """

    def __init__(self, enterprise_configuration):
        """
        Instantiate a new client.

        Args:
            enterprise_configuration (CornerstoneEnterpriseCustomerConfiguration): An enterprise customers's
            configuration model for connecting with Cornerstone
        """
        super().__init__(enterprise_configuration)
        self.global_cornerstone_config = apps.get_model('cornerstone', 'CornerstoneGlobalConfiguration').current()
        self.session = None
        self.expires_at = None

    def create_content_metadata(self, serialized_data):
        """
        Create content metadata using the Cornerstone course content API.
        Since Cornerstone is following pull content model we don't need to implement this method
        """

    def update_content_metadata(self, serialized_data):
        """
        Update content metadata using the Cornerstone course content API.
        Since Cornerstone is following pull content model we don't need to implement this method
        """

    def delete_content_metadata(self, serialized_data):
        """
        Delete content metadata using the Cornerstone course content API.
        Since Cornerstone is following pull content model we don't need to implement this method
        """

    def delete_course_completion(self, user_id, payload):
        """
        Delete a completion status previously sent to the Cornerstone Completion Status endpoint
        Cornerstone does not support this.
        """

    def cleanup_duplicate_assignment_records(self, courses):
        """
        Not implemented yet.
        """
        LOGGER.error(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                "Cornerstone integrated channel does not yet support assignment deduplication."
            )
        )

    def create_course_completion(self, user_id, payload):
        """
        Send a completion status payload to the Cornerstone Completion Status endpoint

        Raises:
            HTTPError: if we received a failure response code from Cornerstone
        """
        json_payload = json.loads(payload)
        callback_url = json_payload['data'].pop('callbackUrl')
        session_token = self.enterprise_configuration.session_token
        if not session_token:
            session_token = json_payload['data'].pop('sessionToken')

        # When exporting content metadata, we encode course keys that contain invalid chars or
        # set them to uuids to comply with Cornerstone standards
        course_id = json_payload['data'].get('courseId')
        key_mapping = get_or_create_key_pair(course_id)
        json_payload['data']['courseId'] = key_mapping.external_course_id
        url = '{base_url}{callback_url}{completion_path}?sessionToken={session_token}'.format(
            base_url=self.enterprise_configuration.cornerstone_base_url,
            callback_url=callback_url,
            completion_path=self.global_cornerstone_config.completion_status_api_path,
            session_token=session_token,
        )

        response = requests.post(
            url,
            json=[json_payload['data']],
            headers={
                'Authorization': self.authorization_header,
                'Content-Type': 'application/json'
            }
        )
        return response.status_code, response.text

    def create_assessment_reporting(self, user_id, payload):
        """
        Not implemented yet
        """

    @property
    def authorization_header(self):
        """
        Authorization header for authenticating requests to cornerstone progress API.
        """
        return 'Basic {}'.format(
            base64.b64encode('{key}:{secret}'.format(
                key=self.global_cornerstone_config.key, secret=self.global_cornerstone_config.secret
            ).encode('utf-8')).decode()
        )



"""
A utility collection for calls from integrated_channels to LMS APIs
If integrated_channels calls LMS APIs, put them here for better tracking.
"""
from opaque_keys.edx.keys import CourseKey

try:
    from lms.djangoapps.certificates.api import get_certificate_for_user
    from lms.djangoapps.grades.course_grade_factory import CourseGradeFactory
    from openedx.core.djangoapps.content.course_overviews.models import CourseOverview
except ImportError:
    get_certificate_for_user = None
    CourseGradeFactory = None
    CourseOverview = None

try:
    from lms.djangoapps.courseware.courses import get_course_blocks_completion_summary
except ImportError:
    get_course_blocks_completion_summary = None

from enterprise.utils import NotConnectedToOpenEdX


def get_course_certificate(course_key, user):
    """
    A course certificate for a user (must be a django.contrib.auth.User instance).
    If there is a problem loading the get_certificate_for_user function, throws NotConnectedToOpenEdX
    If issues with course_key string, throws a InvalidKeyError
    Arguments:
        course_key (string): course key
        user (django.contrib.auth.User): user instance
    Returns a certificate as a dict, for example:
        {
            "username": "bob",
            "course_id": "edX/DemoX/Demo_Course",
            "certificate_type": "verified",
            "created_date": "2015-12-03T13:14:28+0000",
            "status": "downloadable",
            "is_passing": true,
            "download_url": "http://www.example.com/cert.pdf",
            "grade": "0.98"
        }
    """
    if not get_certificate_for_user:
        raise NotConnectedToOpenEdX(
            'To use this function, this package must be '
            'installed in an Open edX environment.'
        )
    course_id = CourseKey.from_string(course_key)
    user_cert = get_certificate_for_user(username=user.username, course_key=course_id)
    return user_cert


def get_single_user_grade(course_key, user):
    """
    Returns a grade for the user (must be a django.contrib.auth.User instance).
    If there is a problem loading the CourseGradeFactory class, throws NotConnectedToOpenEdX
    If issues with course_key string, throws a InvalidKeyError
    Args:
        course_key (string): string course key
        user (django.contrib.auth.User): user instance

    Returns:
        A CourseGrade object with at least these fields:
            - percent (Number)
            - passed (Boolean)
    """
    if not CourseGradeFactory:
        raise NotConnectedToOpenEdX(
            'To use this function, this package must be '
            'installed in an Open edX environment.'
        )
    course_id = CourseKey.from_string(course_key)
    course_grade = CourseGradeFactory().read(user, course_key=course_id)
    return course_grade


def get_course_details(course_key):
    """
    Args:
        course_key (string): string course key
    Returns:
        course_overview: (openedx.core.djangoapps.content.course_overviews.models.CourseOverview)

    If there is a problem loading the CourseOverview class, throws NotConnectedToOpenEdX
    If issues with course_key string, throws a InvalidKeyError
    If course details not found, throws CourseOverview.DoesNotExist
    """
    if not CourseOverview:
        raise NotConnectedToOpenEdX(
            'To use this function, this package must be '
            'installed in an Open edX environment.'
        )
    course_id = CourseKey.from_string(course_key)
    course_overview = CourseOverview.get_from_id(course_id)
    return course_overview


def get_completion_summary(course_key, user):
    """
    Fetch completion summary for course + user using course blocks completions api
    Args:
        course_key (string): string course key
        user (django.contrib.auth.User): user instance
    Returns:
        object containing fields: complete_count, incomplete_count, locked_count

    If there is a problem loading the CourseOverview class, throws NotConnectedToOpenEdX
    If issues with course_key string, throws a InvalidKeyError
    If course details not found, throws CourseOverview.DoesNotExist
    """
    if not get_course_blocks_completion_summary:
        raise NotConnectedToOpenEdX(
            'To use get_course_blocks_completion_summary() function, this package must be '
            'installed in an Open edX environment.'
        )
    course_id = CourseKey.from_string(course_key)
    return get_course_blocks_completion_summary(course_id, user)



"""
Django admin integration for configuring cornerstone ondemand app to communicate with CSOD systems.
"""

from config_models.admin import ConfigurationModelAdmin

from django.apps import apps
from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist

from integrated_channels.cornerstone.models import (
    CornerstoneEnterpriseCustomerConfiguration,
    CornerstoneGlobalConfiguration,
    CornerstoneLearnerDataTransmissionAudit,
)


def enterprise_course_enrollment_model():
    """
    Returns the ``EnterpriseCourseEnrollment`` class.
    """
    return apps.get_model('enterprise', 'EnterpriseCourseEnrollment')


def enterprise_customer_user_model():
    """
    Returns the ``EnterpriseCustomerUser`` class.
    """
    return apps.get_model('enterprise', 'EnterpriseCustomerUser')


def enterprise_customer_model():
    """
    Returns the ``EnterpriseCustomer`` class.
    """
    return apps.get_model('enterprise', 'EnterpriseCustomer')


@admin.register(CornerstoneGlobalConfiguration)
class CornerstoneGlobalConfigurationAdmin(ConfigurationModelAdmin):
    """
    Django admin model for CornerstoneGlobalConfiguration.
    """

    list_display = (
        "completion_status_api_path",
        "key",
        "secret",
    )

    class Meta:
        model = CornerstoneGlobalConfiguration


@admin.register(CornerstoneEnterpriseCustomerConfiguration)
class CornerstoneEnterpriseCustomerConfigurationAdmin(admin.ModelAdmin):
    """
    Django admin model for CornerstoneEnterpriseCustomerConfiguration.
    """

    list_display = (
        "enterprise_customer_name",
        "active",
        "cornerstone_base_url",
    )

    readonly_fields = (
        "enterprise_customer_name",
    )

    list_filter = ("active",)

    search_fields = ("enterprise_customer_name",)

    class Meta:
        model = CornerstoneEnterpriseCustomerConfiguration

    def enterprise_customer_name(self, obj):
        """
        Returns: the name for the attached EnterpriseCustomer.

        Args:
            obj: The instance of CornerstoneEnterpriseCustomerConfiguration
                being rendered with this admin form.
        """
        return obj.enterprise_customer.name


@admin.register(CornerstoneLearnerDataTransmissionAudit)
class CornerstoneLearnerDataTransmissionAuditAdmin(admin.ModelAdmin):
    """
    Django admin model for CornerstoneLearnerDataTransmissionAudit.
    """
    list_display = (
        "user_email",
        "user_id",
        "enterprise_course_enrollment_id",
        "course_id",
        "status",
    )

    readonly_fields = (
        "enterprise_customer_name",
    )

    class Meta:
        model = CornerstoneLearnerDataTransmissionAudit

    def user_email(self, obj):
        """
        Returns: the name for the attached EnterpriseCustomer.

        Args:
            obj: The instance of CornerstoneEnterpriseCustomerConfiguration
                being rendered with this admin form.
        """
        return obj.user.email

    def enterprise_customer_name(self, obj):
        """
        Returns: the name for the attached EnterpriseCustomer.

        Args:
            obj: The instance of CornerstoneEnterpriseCustomerConfiguration
                being rendered with this admin form.
        """

        # a direct foreign key relationship is missing
        # multiple queries here so, avoid adding it to list_display fields
        EnterpriseCourseEnrollment = enterprise_course_enrollment_model()
        EnterpriseCustomerUser = enterprise_customer_user_model()
        EnterpriseCustomer = enterprise_customer_model()
        try:
            ece = EnterpriseCourseEnrollment.objects.get(pk=obj.enterprise_course_enrollment_id)
            ecu = EnterpriseCustomerUser.objects.get(pk=ece.enterprise_customer_user_id)
            ec = EnterpriseCustomer.objects.get(pk=ecu.enterprise_customer_id)
            return ec.name
        except ObjectDoesNotExist:
            return None



"""
Throttles for heartbeat API endpoints.
"""
from rest_framework.throttling import UserRateThrottle


class HeartbeatRateThrottle(UserRateThrottle):
    """
    Throttle class for streamlining heartbeat traffic to manage server load.
    """
    scope = 'heartbeat'
    rate = '60/min'



"""
Django admin integration for configuring moodle app to communicate with Moodle systems.
"""

from django import forms
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from integrated_channels.moodle.models import MoodleEnterpriseCustomerConfiguration


class MoodleEnterpriseCustomerConfigurationForm(forms.ModelForm):
    """
    Django admin form for MoodleEnterpriseCustomerConfiguration.
    """
    class Meta:
        model = MoodleEnterpriseCustomerConfiguration
        fields = '__all__'

    def clean(self):
        cleaned_data = super().clean()
        cleaned_username = cleaned_data.get('username')
        cleaned_password = cleaned_data.get('password')
        cleaned_token = cleaned_data.get('token')
        if cleaned_token and (cleaned_username or cleaned_password):
            raise ValidationError(_('Cannot set both a Username/Password and Token'))
        if (cleaned_username and not cleaned_password) or (cleaned_password and not cleaned_username):
            raise ValidationError(_('Must set both a Username and Password, not just one'))


@admin.register(MoodleEnterpriseCustomerConfiguration)
class MoodleEnterpriseCustomerConfigurationAdmin(admin.ModelAdmin):
    """
    Django admin model for MoodleEnterpriseCustomerConfiguration.
    """

    form = MoodleEnterpriseCustomerConfigurationForm



"""
URL definitions for v1 Integrated Channel API endpoints.
"""

from django.urls import include, path

from .views import IntegratedChannelsBaseViewSet

app_name = 'v1'
urlpatterns = [
    path('canvas/', include('integrated_channels.api.v1.canvas.urls')),
    path('moodle/', include('integrated_channels.api.v1.moodle.urls')),
    path('blackboard/', include('integrated_channels.api.v1.blackboard.urls')),
    path('sap_success_factors/', include('integrated_channels.api.v1.sap_success_factors.urls')),
    path('degreed/', include('integrated_channels.api.v1.degreed.urls')),
    path('degreed2/', include('integrated_channels.api.v1.degreed2.urls')),
    path('cornerstone/', include('integrated_channels.api.v1.cornerstone.urls')),
    path('configs/', IntegratedChannelsBaseViewSet.as_view({'get': 'list'}), name='configs'),
]
